# copyright 2003-2011 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of logilab-common.
#
# logilab-common is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option) any
# later version.
#
# logilab-common is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with logilab-common.  If not, see <http://www.gnu.org/licenses/>.
"""Logilab common library (aka Logilab's extension to the standard library).

:type STD_BLACKLIST: tuple
:var STD_BLACKLIST: directories ignored by default by the functions in
  this package which have to recurse into directories

:type IGNORED_EXTENSIONS: tuple
:var IGNORED_EXTENSIONS: file extensions that may usually be ignored
"""
__docformat__ = "restructuredtext en"

import sys
import types
import pkg_resources
from typing import List, Sequence

__version__ = pkg_resources.get_distribution("logilab-common").version

# deprecated, but keep compatibility with pylint < 1.4.4
__pkginfo__ = types.ModuleType("__pkginfo__")
__pkginfo__.__package__ = __name__
# mypy output: Module has no attribute "version"
# logilab's magic
__pkginfo__.version = __version__  # type: ignore
sys.modules["logilab.common.__pkginfo__"] = __pkginfo__

STD_BLACKLIST = ("CVS", ".svn", ".hg", ".git", ".tox", "debian", "dist", "build")

IGNORED_EXTENSIONS = (".pyc", ".pyo", ".elc", "~", ".swp", ".orig")

# set this to False if you've mx DateTime installed but you don't want your db
# adapter to use it (should be set before you got a connection)
USE_MX_DATETIME = True


class attrdict(dict):
    """A dictionary for which keys are also accessible as attributes."""

    def __getattr__(self, attr: str) -> str:
        try:
            return self[attr]
        except KeyError:
            raise AttributeError(attr)


class dictattr(dict):
    def __init__(self, proxy):
        self.__proxy = proxy

    def __getitem__(self, attr):
        try:
            return getattr(self.__proxy, attr)
        except AttributeError:
            raise KeyError(attr)


class nullobject(object):
    def __repr__(self):
        return "<nullobject>"

    def __bool__(self):
        return False

    __nonzero__ = __bool__


class tempattr(object):
    def __init__(self, obj, attr, value):
        self.obj = obj
        self.attr = attr
        self.value = value

    def __enter__(self):
        self.oldvalue = getattr(self.obj, self.attr)
        setattr(self.obj, self.attr, self.value)
        return self.obj

    def __exit__(self, exctype, value, traceback):
        setattr(self.obj, self.attr, self.oldvalue)


# flatten -----
# XXX move in a specific module and use yield instead
# do not mix flatten and translate
#
# def iterable(obj):
#    try: iter(obj)
#    except: return False
#    return True
#
# def is_string_like(obj):
#    try: obj +''
#    except (TypeError, ValueError): return False
#    return True
#
# def is_scalar(obj):
#    return is_string_like(obj) or not iterable(obj)
#
# def flatten(seq):
#    for item in seq:
#        if is_scalar(item):
#            yield item
#        else:
#            for subitem in flatten(item):
#               yield subitem


def flatten(iterable, tr_func=None, results=None):
    """Flatten a list of list with any level.

    If tr_func is not None, it should be a one argument function that'll be called
    on each final element.

    :rtype: list

    >>> flatten([1, [2, 3]])
    [1, 2, 3]
    """
    if results is None:
        results = []
    for val in iterable:
        if isinstance(val, (list, tuple)):
            flatten(val, tr_func, results)
        elif tr_func is None:
            results.append(val)
        else:
            results.append(tr_func(val))
    return results


# XXX is function below still used ?


def make_domains(lists):
    """
    Given a list of lists, return a list of domain for each list to produce all
    combinations of possibles values.

    :rtype: list

    Example:

    >>> make_domains(['a', 'b'], ['c','d', 'e'])
    [['a', 'b', 'a', 'b', 'a', 'b'], ['c', 'c', 'd', 'd', 'e', 'e']]
    """
    domains = []
    for iterable in lists:
        new_domain = iterable[:]
        for i in range(len(domains)):
            domains[i] = domains[i] * len(iterable)
        if domains:
            missing = (len(domains[0]) - len(iterable)) / len(iterable)
            i = 0
            for j in range(len(iterable)):
                value = iterable[j]
                for dummy in range(missing):
                    new_domain.insert(i, value)
                    i += 1
                i += 1
        domains.append(new_domain)
    return domains


# private stuff ################################################################


def _handle_blacklist(blacklist: Sequence[str], dirnames: List[str], filenames: List[str]) -> None:
    """remove files/directories in the black list

    dirnames/filenames are usually from os.walk
    """
    for norecurs in blacklist:
        if norecurs in dirnames:
            dirnames.remove(norecurs)
        elif norecurs in filenames:
            filenames.remove(norecurs)



# copyright 2003-2011 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of logilab-common.
#
# logilab-common is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option) any
# later version.
#
# logilab-common is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with logilab-common.  If not, see <http://www.gnu.org/licenses/>.
"""logilab-pytest is a tool that eases test running and debugging.

To be able to use logilab-pytest, you should either write tests using
the logilab.common.testlib's framework or the unittest module of the
Python's standard library.

You can customize logilab-pytest's behaviour by defining a ``pytestconf.py``
file somewhere in your test directory. In this file, you can add options or
change the way tests are run.

To add command line options, you must define a ``update_parser`` function in
your ``pytestconf.py`` file. The function must accept a single parameter
that will be the OptionParser's instance to customize.

If you wish to customize the tester, you'll have to define a class named
``CustomPyTester``. This class should extend the default `PyTester` class
defined in the logilab.common.pytest module. Take a look at the `PyTester` and
`DjangoTester` classes for more information about what can be done.

For instance, if you wish to add a custom -l option to specify a loglevel, you
could define the following ``pytestconf.py`` file ::

    import logging
    from logilab.common.pytest import PyTester

    def update_parser(parser):
        parser.add_option('-l', '--loglevel', dest='loglevel', action='store',
                          choices=('debug', 'info', 'warning', 'error', 'critical'),
                          default='critical', help="the default log level possible choices are "
                          "('debug', 'info', 'warning', 'error', 'critical')")
        return parser


    class CustomPyTester(PyTester):
        def __init__(self, cvg, options):
            super(CustomPyTester, self).__init__(cvg, options)
            loglevel = options.loglevel.upper()
            logger = logging.getLogger('erudi')
            logger.setLevel(logging.getLevelName(loglevel))


In your TestCase class you can then get the value of a specific option with
the ``optval`` method::

    class MyTestCase(TestCase):
        def test_foo(self):
            loglevel = self.optval('loglevel')
            # ...


You can also tag your tag your test for fine filtering

With those tag::

    from logilab.common.testlib import tag, TestCase

    class Exemple(TestCase):

        @tag('rouge', 'carre')
        def toto(self):
            pass

        @tag('carre', 'vert')
        def tata(self):
            pass

        @tag('rouge')
        def titi(test):
            pass

you can filter the function with a simple python expression

 * ``toto`` and ``titi`` match ``rouge``
 * ``toto``, ``tata`` and ``titi``, match ``rouge or carre``
 * ``tata`` and ``titi`` match``rouge ^ carre``
 * ``titi`` match ``rouge and not carre``
"""

from __future__ import print_function

import os
import re
import sys
import os.path as osp
from time import process_time, time

try:
    from re import Match
except ImportError:
    # Match is python > 3.6 only.
    #
    # To be compatible with python <= 3.6, and still provide some typing, we
    # manually define Match, in the same manner it is defined in the re module
    # of python > 3.7
    # cf https://github.com/python/cpython/blob/3.7/Lib/re.py#L264
    Match = type(re.sre_compile.compile("", 0).match(""))  # type: ignore
import warnings
import types
import inspect
import traceback
from inspect import isgeneratorfunction, isclass, FrameInfo
from random import shuffle
from itertools import dropwhile

# mypy error: Module 'unittest.runner' has no attribute '_WritelnDecorator'
# but it does
from unittest.runner import _WritelnDecorator  # type: ignore
from unittest.suite import TestSuite

from typing import Callable, Any, Optional, List, Tuple, Generator, Dict

from logilab.common.deprecation import callable_deprecated
from logilab.common.fileutils import abspath_listdir
from logilab.common import textutils
from logilab.common import testlib, STD_BLACKLIST

# use the same unittest module as testlib
from logilab.common.testlib import unittest, start_interactive_mode
from logilab.common.testlib import (  # noqa
    nocoverage,
    pause_trace,
    replace_trace,
)  # bwcompat
from logilab.common.debugger import Debugger, colorize_source
import doctest

import unittest as unittest_legacy

from .decorators import monkeypatch

__docformat__ = "restructuredtext en"

PYTEST_DOC = """%prog [OPTIONS] [testfile [testpattern]]

examples:

logilab-pytest path/to/mytests.py
logilab-pytest path/to/mytests.py TheseTests
logilab-pytest path/to/mytests.py TheseTests.test_thisone
logilab-pytest path/to/mytests.py -m '(not long and database) or regr'

logilab-pytest one (will run both test_thisone and test_thatone)
logilab-pytest path/to/mytests.py -s not (will skip test_notthisone)
"""

ENABLE_DBC = False
FILE_RESTART = ".pytest.restart"

if not getattr(unittest_legacy, "__package__", None):
    try:
        import unittest2.suite as unittest_suite
    except ImportError:
        sys.exit("You have to install python-unittest2 to use this module")
else:
    # mypy: Name 'unittest_suite' already defined (possibly by an import))
    import unittest.suite as unittest_suite  # type: ignore

try:
    import django  # noqa
    from logilab.common.modutils import modpath_from_file, load_module_from_modpath

    DJANGO_FOUND = True
except ImportError:
    DJANGO_FOUND = False

CONF_FILE = "pytestconf.py"

TESTFILE_RE = re.compile(r"^((unit)?test.*|smoketest)\.py$")


def this_is_a_testfile(filename: str) -> Optional[Match]:
    """returns True if `filename` seems to be a test file"""
    return TESTFILE_RE.match(osp.basename(filename))


TESTDIR_RE = re.compile("^(unit)?tests?$")


def this_is_a_testdir(dirpath: str) -> Optional[Match]:
    """returns True if `filename` seems to be a test directory"""
    return TESTDIR_RE.match(osp.basename(dirpath))


def load_pytest_conf(path, parser):
    """loads a ``pytestconf.py`` file and update default parser
    and / or tester.
    """
    namespace = {}
    exec(open(path, "rb").read(), namespace)
    if "update_parser" in namespace:
        namespace["update_parser"](parser)
    return namespace.get("CustomPyTester", PyTester)


def project_root(parser, projdir=None):
    """try to find project's root and add it to sys.path"""
    if projdir is None:
        projdir = os.getcwd()
    previousdir = curdir = osp.abspath(projdir)
    testercls = PyTester
    conf_file_path = osp.join(curdir, CONF_FILE)
    if osp.isfile(conf_file_path):
        testercls = load_pytest_conf(conf_file_path, parser)
    while this_is_a_testdir(curdir) or osp.isfile(osp.join(curdir, "__init__.py")):
        newdir = osp.normpath(osp.join(curdir, os.pardir))
        if newdir == curdir:
            break
        previousdir = curdir
        curdir = newdir
        conf_file_path = osp.join(curdir, CONF_FILE)
        if osp.isfile(conf_file_path):
            testercls = load_pytest_conf(conf_file_path, parser)
    return previousdir, testercls


class GlobalTestReport(object):
    """this class holds global test statistics"""

    def __init__(self):
        self.ran = 0
        self.skipped = 0
        self.failures = 0
        self.errors = 0
        self.ttime = 0
        self.ctime = 0
        self.modulescount = 0
        self.errmodules = []

    def feed(self, filename, testresult, ttime, ctime):
        """integrates new test information into internal statistics"""
        ran = testresult.testsRun
        self.ran += ran
        self.skipped += len(getattr(testresult, "skipped", ()))
        self.failures += len(testresult.failures)
        self.errors += len(testresult.errors)
        self.ttime += ttime
        self.ctime += ctime
        self.modulescount += 1
        if not testresult.wasSuccessful():
            problems = len(testresult.failures) + len(testresult.errors)
            self.errmodules.append((filename[:-3], problems, ran))

    def failed_to_test_module(self, filename):
        """called when the test module could not be imported by unittest"""
        self.errors += 1
        self.modulescount += 1
        self.ran += 1
        self.errmodules.append((filename[:-3], 1, 1))

    def skip_module(self, filename):
        self.modulescount += 1
        self.ran += 1
        self.errmodules.append((filename[:-3], 0, 0))

    def __str__(self):
        """this is just presentation stuff"""
        line1 = ["Ran %s test cases in %.2fs (%.2fs CPU)" % (self.ran, self.ttime, self.ctime)]
        if self.errors:
            line1.append("%s errors" % self.errors)
        if self.failures:
            line1.append("%s failures" % self.failures)
        if self.skipped:
            line1.append("%s skipped" % self.skipped)
        modulesok = self.modulescount - len(self.errmodules)
        if self.errors or self.failures:
            line2 = "%s modules OK (%s failed)" % (modulesok, len(self.errmodules))
            descr = ", ".join(["%s [%s/%s]" % info for info in self.errmodules])
            line3 = "\nfailures: %s" % descr
        elif modulesok:
            line2 = "All %s modules OK" % modulesok
            line3 = ""
        else:
            return ""
        return "%s\n%s%s" % (", ".join(line1), line2, line3)


def remove_local_modules_from_sys(testdir):
    """remove all modules from cache that come from `testdir`

    This is used to avoid strange side-effects when using the
    testall() mode of pytest.
    For instance, if we run pytest on this tree::

      A/test/test_utils.py
      B/test/test_utils.py

    we **have** to clean sys.modules to make sure the correct test_utils
    module is ran in B
    """
    for modname, mod in list(sys.modules.items()):
        if mod is None:
            continue
        if not hasattr(mod, "__file__"):
            # this is the case of some built-in modules like sys, imp, marshal
            continue
        modfile = mod.__file__
        # if modfile is not an absolute path, it was probably loaded locally
        # during the tests
        if not osp.isabs(modfile) or modfile.startswith(testdir):
            del sys.modules[modname]


class PyTester(object):
    """encapsulates testrun logic"""

    def __init__(self, cvg, options):
        self.report = GlobalTestReport()
        self.cvg = cvg
        self.options = options
        self.firstwrite = True
        self._errcode = None

    def show_report(self):
        """prints the report and returns appropriate exitcode"""
        # everything has been ran, print report
        print("*" * 79)
        print(self.report)

    def get_errcode(self):
        # errcode set explicitly
        if self._errcode is not None:
            return self._errcode
        return self.report.failures + self.report.errors

    def set_errcode(self, errcode):
        self._errcode = errcode

    errcode = property(get_errcode, set_errcode)

    def testall(self, exitfirst=False):
        """walks through current working directory, finds something
        which can be considered as a testdir and runs every test there
        """
        here = os.getcwd()
        for dirname, dirs, _ in os.walk(here):
            for skipped in STD_BLACKLIST:
                if skipped in dirs:
                    dirs.remove(skipped)
            basename = osp.basename(dirname)
            if this_is_a_testdir(basename):
                print("going into", dirname)
                # we found a testdir, let's explore it !
                if not self.testonedir(dirname, exitfirst):
                    break
                dirs[:] = []
        if self.report.ran == 0:
            print("no test dir found testing here:", here)
            # if no test was found during the visit, consider
            # the local directory as a test directory even if
            # it doesn't have a traditional test directory name
            self.testonedir(here)

    def testonedir(self, testdir, exitfirst=False):
        """finds each testfile in the `testdir` and runs it

        return true when all tests has been executed, false if exitfirst and
        some test has failed.
        """
        files = abspath_listdir(testdir)
        shuffle(files)
        for filename in files:
            if this_is_a_testfile(filename):
                if self.options.exitfirst and not self.options.restart:
                    # overwrite restart file
                    try:
                        restartfile = open(FILE_RESTART, "w")
                        restartfile.close()
                    except Exception:
                        print(
                            "Error while overwriting succeeded test file :",
                            osp.join(os.getcwd(), FILE_RESTART),
                            file=sys.__stderr__,
                        )
                        raise
                # run test and collect information
                prog = self.testfile(filename, batchmode=True)
                if exitfirst and (prog is None or not prog.result.wasSuccessful()):
                    return False
                self.firstwrite = True
        # clean local modules
        remove_local_modules_from_sys(testdir)
        return True

    def testfile(self, filename, batchmode=False):
        """runs every test in `filename`

        :param filename: an absolute path pointing to a unittest file
        """
        here = os.getcwd()
        dirname = osp.dirname(filename)
        if dirname:
            os.chdir(dirname)
        # overwrite restart file if it has not been done already
        if self.options.exitfirst and not self.options.restart and self.firstwrite:
            try:
                restartfile = open(FILE_RESTART, "w")
                restartfile.close()
            except Exception:
                print(
                    "Error while overwriting succeeded test file :",
                    osp.join(os.getcwd(), FILE_RESTART),
                    file=sys.__stderr__,
                )
                raise
        modname = osp.basename(filename)[:-3]
        print(("  %s  " % osp.basename(filename)).center(70, "="), file=sys.__stderr__)
        try:
            tstart, cstart = time(), process_time()
            try:
                testprog = SkipAwareTestProgram(
                    modname,
                    batchmode=batchmode,
                    cvg=self.cvg,
                    options=self.options,
                    outstream=sys.stderr,
                )
            except KeyboardInterrupt:
                raise
            except SystemExit as exc:
                self.errcode = exc.code
                raise
            except testlib.SkipTest:
                print("Module skipped:", filename)
                self.report.skip_module(filename)
                return None
            except Exception:
                self.report.failed_to_test_module(filename)
                print("unhandled exception occurred while testing", modname, file=sys.stderr)
                import traceback

                traceback.print_exc(file=sys.stderr)
                return None

            tend, cend = time(), process_time()
            ttime, ctime = (tend - tstart), (cend - cstart)
            self.report.feed(filename, testprog.result, ttime, ctime)
            return testprog
        finally:
            if dirname:
                os.chdir(here)


class DjangoTester(PyTester):
    def load_django_settings(self, dirname):
        """try to find project's setting and load it"""
        curdir = osp.abspath(dirname)
        while not osp.isfile(osp.join(curdir, "settings.py")) and osp.isfile(
            osp.join(curdir, "__init__.py")
        ):
            newdir = osp.normpath(osp.join(curdir, os.pardir))
            if newdir == curdir:
                raise AssertionError("could not find settings.py")
            curdir = newdir
        # late django initialization
        settings = load_module_from_modpath(modpath_from_file(osp.join(curdir, "settings.py")))
        from django.core.management import setup_environ

        setup_environ(settings)
        settings.DEBUG = False
        self.settings = settings
        # add settings dir to pythonpath since it's the project's root
        if curdir not in sys.path:
            sys.path.insert(1, curdir)

    def before_testfile(self):
        # Those imports must be done **after** setup_environ was called
        from django.test.utils import setup_test_environment
        from django.test.utils import create_test_db

        setup_test_environment()
        create_test_db(verbosity=0)
        self.dbname = self.settings.TEST_DATABASE_NAME

    def after_testfile(self):
        # Those imports must be done **after** setup_environ was called
        from django.test.utils import teardown_test_environment
        from django.test.utils import destroy_test_db

        teardown_test_environment()
        print("destroying", self.dbname)
        destroy_test_db(self.dbname, verbosity=0)

    def testall(self, exitfirst=False):
        """walks through current working directory, finds something
        which can be considered as a testdir and runs every test there
        """
        for dirname, dirs, files in os.walk(os.getcwd()):
            for skipped in ("CVS", ".svn", ".hg"):
                if skipped in dirs:
                    dirs.remove(skipped)
            if "tests.py" in files:
                if not self.testonedir(dirname, exitfirst):
                    break
                dirs[:] = []
            else:
                basename = osp.basename(dirname)
                if basename in ("test", "tests"):
                    print("going into", dirname)
                    # we found a testdir, let's explore it !
                    if not self.testonedir(dirname, exitfirst):
                        break
                    dirs[:] = []

    def testonedir(self, testdir, exitfirst=False):
        """finds each testfile in the `testdir` and runs it

        return true when all tests has been executed, false if exitfirst and
        some test has failed.
        """
        # special django behaviour : if tests are splitted in several files,
        # remove the main tests.py file and tests each test file separately
        testfiles = [fpath for fpath in abspath_listdir(testdir) if this_is_a_testfile(fpath)]
        if len(testfiles) > 1:
            try:
                testfiles.remove(osp.join(testdir, "tests.py"))
            except ValueError:
                pass
        for filename in testfiles:
            # run test and collect information
            prog = self.testfile(filename, batchmode=True)
            if exitfirst and (prog is None or not prog.result.wasSuccessful()):
                return False
        # clean local modules
        remove_local_modules_from_sys(testdir)
        return True

    def testfile(self, filename, batchmode=False):
        """runs every test in `filename`

        :param filename: an absolute path pointing to a unittest file
        """
        here = os.getcwd()
        dirname = osp.dirname(filename)
        if dirname:
            os.chdir(dirname)
        self.load_django_settings(dirname)
        modname = osp.basename(filename)[:-3]
        print(("  %s  " % osp.basename(filename)).center(70, "="), file=sys.stderr)
        try:
            try:
                tstart, cstart = time(), process_time()
                self.before_testfile()
                testprog = SkipAwareTestProgram(modname, batchmode=batchmode, cvg=self.cvg)
                tend, cend = time(), process_time()
                ttime, ctime = (tend - tstart), (cend - cstart)
                self.report.feed(filename, testprog.result, ttime, ctime)
                return testprog
            except SystemExit:
                raise
            except Exception as exc:
                import traceback

                traceback.print_exc()
                self.report.failed_to_test_module(filename)
                print("unhandled exception occurred while testing", modname)
                print("error: %s" % exc)
                return None
        finally:
            self.after_testfile()
            if dirname:
                os.chdir(here)


def make_parser():
    """creates the OptionParser instance"""
    from optparse import OptionParser

    parser = OptionParser(usage=PYTEST_DOC)

    parser.newargs = []

    def rebuild_cmdline(option, opt, value, parser):
        """carry the option to unittest_main"""
        parser.newargs.append(opt)

    def rebuild_and_store(option, opt, value, parser):
        """carry the option to unittest_main and store
        the value on current parser
        """
        parser.newargs.append(opt)
        setattr(parser.values, option.dest, True)

    def capture_and_rebuild(option, opt, value, parser):
        warnings.simplefilter("ignore", DeprecationWarning)
        rebuild_cmdline(option, opt, value, parser)

    # logilab-pytest options
    parser.add_option(
        "-t", dest="testdir", default=None, help="directory where the tests will be found"
    )
    parser.add_option(
        "-d", dest="dbc", default=False, action="store_true", help="enable design-by-contract"
    )
    # unittest_main options provided and passed through logilab-pytest
    parser.add_option(
        "-v", "--verbose", callback=rebuild_cmdline, action="callback", help="Verbose output"
    )
    parser.add_option(
        "-i",
        "--pdb",
        callback=rebuild_and_store,
        dest="pdb",
        action="callback",
        help="Enable test failure inspection",
    )
    parser.add_option(
        "-x",
        "--exitfirst",
        callback=rebuild_and_store,
        dest="exitfirst",
        default=False,
        action="callback",
        help="Exit on first failure " "(only make sense when logilab-pytest run one test file)",
    )
    parser.add_option(
        "-R",
        "--restart",
        callback=rebuild_and_store,
        dest="restart",
        default=False,
        action="callback",
        help="Restart tests from where it failed (implies exitfirst) "
        "(only make sense if tests previously ran with exitfirst only)",
    )
    parser.add_option(
        "--color", callback=rebuild_cmdline, action="callback", help="colorize tracebacks"
    )
    parser.add_option(
        "-s",
        "--skip",
        # XXX: I wish I could use the callback action but it
        #      doesn't seem to be able to get the value
        #      associated to the option
        action="store",
        dest="skipped",
        default=None,
        help="test names matching this name will be skipped "
        "to skip several patterns, use commas",
    )
    parser.add_option(
        "-q", "--quiet", callback=rebuild_cmdline, action="callback", help="Minimal output"
    )
    parser.add_option(
        "-P",
        "--profile",
        default=None,
        dest="profile",
        help="Profile execution and store data in the given file",
    )
    parser.add_option(
        "-m",
        "--match",
        default=None,
        dest="tags_pattern",
        help="only execute test whose tag match the current pattern",
    )

    if DJANGO_FOUND:
        parser.add_option(
            "-J",
            "--django",
            dest="django",
            default=False,
            action="store_true",
            help="use logilab-pytest for django test cases",
        )
    return parser


def parseargs(parser):
    """Parse the command line and return (options processed), (options to pass to
    unittest_main()), (explicitfile or None).
    """
    # parse the command line
    options, args = parser.parse_args()
    filenames = [arg for arg in args if arg.endswith(".py")]
    if filenames:
        if len(filenames) > 1:
            parser.error("only one filename is acceptable")
        explicitfile = filenames[0]
        args.remove(explicitfile)
    else:
        explicitfile = None
    # someone wants DBC
    testlib.ENABLE_DBC = options.dbc
    newargs = parser.newargs
    if options.skipped:
        newargs.extend(["--skip", options.skipped])
    # restart implies exitfirst
    if options.restart:
        options.exitfirst = True
    # append additional args to the new sys.argv and let unittest_main
    # do the rest
    newargs += args
    return options, explicitfile


@callable_deprecated("[logilab-common 1.3] logilab-pytest is deprecated, use another test runner")
def run():
    parser = make_parser()
    rootdir, testercls = project_root(parser)
    options, explicitfile = parseargs(parser)
    # mock a new command line
    sys.argv[1:] = parser.newargs
    cvg = None
    if "" not in sys.path:
        sys.path.insert(0, "")
    if DJANGO_FOUND and options.django:
        tester = DjangoTester(cvg, options)
    else:
        tester = testercls(cvg, options)
    if explicitfile:
        cmd, args = tester.testfile, (explicitfile,)
    elif options.testdir:
        cmd, args = tester.testonedir, (options.testdir, options.exitfirst)
    else:
        cmd, args = tester.testall, (options.exitfirst,)
    try:
        try:
            if options.profile:
                import hotshot

                prof = hotshot.Profile(options.profile)
                prof.runcall(cmd, *args)
                prof.close()
                print("profile data saved in", options.profile)
            else:
                cmd(*args)
        except SystemExit:
            raise
        except Exception:
            import traceback

            traceback.print_exc()
    finally:
        tester.show_report()
        sys.exit(tester.errcode)


class SkipAwareTestProgram(unittest.TestProgram):
    # XXX: don't try to stay close to unittest.py, use optparse
    USAGE = """\
Usage: %(progName)s [options] [test] [...]

Options:
  -h, --help       Show this message
  -v, --verbose    Verbose output
  -i, --pdb        Enable test failure inspection
  -x, --exitfirst  Exit on first failure
  -s, --skip       skip test matching this pattern (no regexp for now)
  -q, --quiet      Minimal output
  --color          colorize tracebacks

  -m, --match      Run only test whose tag match this pattern

  -P, --profile    FILE: Run the tests using cProfile and saving results
                   in FILE

Examples:
  %(progName)s                               - run default set of tests
  %(progName)s MyTestSuite                   - run suite 'MyTestSuite'
  %(progName)s MyTestCase.testSomething      - run MyTestCase.testSomething
  %(progName)s MyTestCase                    - run all 'test*' test methods
                                               in MyTestCase
"""

    def __init__(
        self,
        module="__main__",
        defaultTest=None,
        batchmode=False,
        cvg=None,
        options=None,
        outstream=sys.stderr,
    ):
        self.batchmode = batchmode
        self.cvg = cvg
        self.options = options
        self.outstream = outstream
        super(SkipAwareTestProgram, self).__init__(
            module=module, defaultTest=defaultTest, testLoader=NonStrictTestLoader()
        )

    def parseArgs(self, argv):
        self.pdbmode = False
        self.exitfirst = False
        self.skipped_patterns = []
        self.test_pattern = None
        self.tags_pattern = None
        self.colorize = False
        self.profile_name = None
        import getopt

        try:
            options, args = getopt.getopt(
                argv[1:],
                "hHvixrqcp:s:m:P:",
                [
                    "help",
                    "verbose",
                    "quiet",
                    "pdb",
                    "exitfirst",
                    "restart",
                    "skip=",
                    "color",
                    "match=",
                    "profile=",
                ],
            )
            for opt, value in options:
                if opt in ("-h", "-H", "--help"):
                    self.usageExit()
                if opt in ("-i", "--pdb"):
                    self.pdbmode = True
                if opt in ("-x", "--exitfirst"):
                    self.exitfirst = True
                if opt in ("-r", "--restart"):
                    self.restart = True
                    self.exitfirst = True
                if opt in ("-q", "--quiet"):
                    self.verbosity = 0
                if opt in ("-v", "--verbose"):
                    self.verbosity = 2
                if opt in ("-s", "--skip"):
                    self.skipped_patterns = [pat.strip() for pat in value.split(", ")]
                if opt == "--color":
                    self.colorize = True
                if opt in ("-m", "--match"):
                    # self.tags_pattern = value
                    self.options["tag_pattern"] = value
                if opt in ("-P", "--profile"):
                    self.profile_name = value
            self.testLoader.skipped_patterns = self.skipped_patterns
            if len(args) == 0 and self.defaultTest is None:
                suitefunc = getattr(self.module, "suite", None)
                if isinstance(suitefunc, (types.FunctionType, types.MethodType)):
                    self.test = self.module.suite()
                else:
                    self.test = self.testLoader.loadTestsFromModule(self.module)
                return
            if len(args) > 0:
                self.test_pattern = args[0]
                self.testNames = args
            else:
                self.testNames = (self.defaultTest,)
            self.createTests()
        except getopt.error as msg:
            self.usageExit(msg)

    def runTests(self):
        if self.profile_name:
            import cProfile

            cProfile.runctx("self._runTests()", globals(), locals(), self.profile_name)
        else:
            return self._runTests()

    def _runTests(self):
        self.testRunner = SkipAwareTextTestRunner(
            verbosity=self.verbosity,
            stream=self.outstream,
            exitfirst=self.exitfirst,
            pdbmode=self.pdbmode,
            cvg=self.cvg,
            test_pattern=self.test_pattern,
            skipped_patterns=self.skipped_patterns,
            colorize=self.colorize,
            batchmode=self.batchmode,
            options=self.options,
        )

        def removeSucceededTests(obj, succTests):
            """Recursive function that removes succTests from
            a TestSuite or TestCase
            """
            if isinstance(obj, unittest.TestSuite):
                removeSucceededTests(obj._tests, succTests)
            if isinstance(obj, list):
                for el in obj[:]:
                    if isinstance(el, unittest.TestSuite):
                        removeSucceededTests(el, succTests)
                    elif isinstance(el, unittest.TestCase):
                        descr = ".".join(
                            (el.__class__.__module__, el.__class__.__name__, el._testMethodName)
                        )
                        if descr in succTests:
                            obj.remove(el)

        # take care, self.options may be None
        if getattr(self.options, "restart", False):
            # retrieve succeeded tests from FILE_RESTART
            try:
                restartfile = open(FILE_RESTART, "r")
                try:
                    succeededtests = list(elem.rstrip("\n\r") for elem in restartfile.readlines())
                    removeSucceededTests(self.test, succeededtests)
                finally:
                    restartfile.close()
            except Exception as ex:
                raise Exception(
                    "Error while reading succeeded tests into %s: %s"
                    % (osp.join(os.getcwd(), FILE_RESTART), ex)
                )

        result = self.testRunner.run(self.test)
        # help garbage collection: we want TestSuite, which hold refs to every
        # executed TestCase, to be gc'ed
        del self.test
        if getattr(result, "debuggers", None) and getattr(self, "pdbmode", None):
            start_interactive_mode(result)
        if not getattr(self, "batchmode", None):
            sys.exit(not result.wasSuccessful())
        self.result = result


class SkipAwareTextTestRunner(unittest.TextTestRunner):
    def __init__(
        self,
        stream=sys.stderr,
        verbosity=1,
        exitfirst=False,
        pdbmode=False,
        cvg=None,
        test_pattern=None,
        skipped_patterns=(),
        colorize=False,
        batchmode=False,
        options=None,
    ):
        super(SkipAwareTextTestRunner, self).__init__(stream=stream, verbosity=verbosity)
        self.exitfirst = exitfirst
        self.pdbmode = pdbmode
        self.cvg = cvg
        self.test_pattern = test_pattern
        self.skipped_patterns = skipped_patterns
        self.colorize = colorize
        self.batchmode = batchmode
        self.options = options

    def _this_is_skipped(self, testedname: str) -> bool:
        return any([(pat in testedname) for pat in self.skipped_patterns])

    def _runcondition(self, test: Callable, skipgenerator: bool = True) -> bool:
        if isinstance(test, testlib.InnerTest):
            testname = test.name
        else:
            if isinstance(test, testlib.TestCase):
                meth = test._get_test_method()
                testname = "%s.%s" % (test.__name__, meth.__name__)
            elif isinstance(test, types.FunctionType):
                func = test
                testname = func.__name__
            elif isinstance(test, types.MethodType):
                cls = test.__self__.__class__
                testname = "%s.%s" % (cls.__name__, test.__name__)
            else:
                return True  # Not sure when this happens
            if isgeneratorfunction(test) and skipgenerator:
                return self.does_match_tags(test)  # Let inner tests decide at run time
        if self._this_is_skipped(testname):
            return False  # this was explicitly skipped
        if self.test_pattern is not None:
            try:
                classpattern, testpattern = self.test_pattern.split(".")
                klass, name = testname.split(".")
                if classpattern not in klass or testpattern not in name:
                    return False
            except ValueError:
                if self.test_pattern not in testname:
                    return False

        return self.does_match_tags(test)

    def does_match_tags(self, test: Callable) -> bool:
        if self.options is not None:
            tags_pattern = getattr(self.options, "tags_pattern", None)
            if tags_pattern is not None:
                tags = getattr(test, "tags", testlib.Tags())
                if tags.inherit and isinstance(test, types.MethodType):
                    tags = tags | getattr(test.__self__.__class__, "tags", testlib.Tags())
                return tags.match(tags_pattern)
        return True  # no pattern

    def _makeResult(self) -> "SkipAwareTestResult":
        return SkipAwareTestResult(
            self.stream,
            self.descriptions,
            self.verbosity,
            self.exitfirst,
            self.pdbmode,
            self.cvg,
            self.colorize,
        )

    def run(self, test):
        "Run the given test case or test suite."
        result = self._makeResult()
        startTime = time()
        test(result, runcondition=self._runcondition, options=self.options)
        stopTime = time()
        timeTaken = stopTime - startTime
        result.printErrors()
        if not self.batchmode:
            self.stream.writeln(result.separator2)
            run = result.testsRun
            self.stream.writeln("Ran %d test%s in %.3fs" % (run, run != 1 and "s" or "", timeTaken))
            self.stream.writeln()
            if not result.wasSuccessful():
                if self.colorize:
                    self.stream.write(textutils.colorize_ansi("FAILED", color="red"))
                else:
                    self.stream.write("FAILED")
            else:
                if self.colorize:
                    self.stream.write(textutils.colorize_ansi("OK", color="green"))
                else:
                    self.stream.write("OK")
            failed, errored, skipped = map(len, (result.failures, result.errors, result.skipped))

            det_results = []
            for name, value in (
                ("failures", result.failures),
                ("errors", result.errors),
                ("skipped", result.skipped),
            ):
                if value:
                    det_results.append("%s=%i" % (name, len(value)))
            if det_results:
                self.stream.write(" (")
                self.stream.write(", ".join(det_results))
                self.stream.write(")")
            self.stream.writeln("")
        return result


class SkipAwareTestResult(unittest._TextTestResult):
    def __init__(
        self,
        stream: _WritelnDecorator,
        descriptions: bool,
        verbosity: int,
        exitfirst: bool = False,
        pdbmode: bool = False,
        cvg: Optional[Any] = None,
        colorize: bool = False,
    ) -> None:
        super(SkipAwareTestResult, self).__init__(stream, descriptions, verbosity)
        self.skipped: List[Tuple[Any, Any]] = []
        self.debuggers: List = []
        self.fail_descrs: List = []
        self.error_descrs: List = []
        self.exitfirst = exitfirst
        self.pdbmode = pdbmode
        self.cvg = cvg
        self.colorize = colorize
        self.pdbclass = Debugger
        self.verbose = verbosity > 1

    def descrs_for(self, flavour: str) -> List[Tuple[int, str]]:
        return getattr(self, "%s_descrs" % flavour.lower())

    def _create_pdb(self, test_descr: str, flavour: str) -> None:
        self.descrs_for(flavour).append((len(self.debuggers), test_descr))
        if self.pdbmode:
            self.debuggers.append(self.pdbclass(sys.exc_info()[2]))

    def _iter_valid_frames(self, frames: List[FrameInfo]) -> Generator[FrameInfo, Any, None]:
        """only consider non-testlib frames when formatting  traceback"""

        def invalid(fi):
            return osp.abspath(fi[1]) in (lgc_testlib, std_testlib)

        lgc_testlib = osp.abspath(__file__)
        std_testlib = osp.abspath(unittest.__file__)

        for frameinfo in dropwhile(invalid, frames):
            yield frameinfo

    def _exc_info_to_string(self, err, test):
        """Converts a sys.exc_info()-style tuple of values into a string.

        This method is overridden here because we want to colorize
        lines if --color is passed, and display local variables if
        --verbose is passed
        """
        exctype, exc, tb = err
        output = ["Traceback (most recent call last)"]
        frames = inspect.getinnerframes(tb)
        colorize = self.colorize
        frames = enumerate(self._iter_valid_frames(frames))
        for index, (frame, filename, lineno, funcname, ctx, ctxindex) in frames:
            filename = osp.abspath(filename)
            if ctx is None:  # pyc files or C extensions for instance
                source = "<no source available>"
            else:
                source = "".join(ctx)
            if colorize:
                filename = textutils.colorize_ansi(filename, "magenta")
                source = colorize_source(source)
            output.append('  File "%s", line %s, in %s' % (filename, lineno, funcname))
            output.append("    %s" % source.strip())
            if self.verbose:
                output.append("%r == %r" % (dir(frame), test.__module__))
                output.append("")
                output.append("    " + " local variables ".center(66, "-"))
                for varname, value in sorted(frame.f_locals.items()):
                    output.append("    %s: %r" % (varname, value))
                    if varname == "self":  # special handy processing for self
                        for varname, value in sorted(vars(value).items()):
                            output.append("      self.%s: %r" % (varname, value))
                output.append("    " + "-" * 66)
                output.append("")
        output.append("".join(traceback.format_exception_only(exctype, exc)))
        return "\n".join(output)

    def addError(self, test, err):
        """err ->  (exc_type, exc, tcbk)"""
        exc_type, exc, _ = err
        if isinstance(exc, testlib.SkipTest):
            assert exc_type == testlib.SkipTest
            self.addSkip(test, exc)
        else:
            if self.exitfirst:
                self.shouldStop = True
            descr = self.getDescription(test)
            super(SkipAwareTestResult, self).addError(test, err)
            self._create_pdb(descr, "error")

    def addFailure(self, test, err):
        if self.exitfirst:
            self.shouldStop = True
        descr = self.getDescription(test)
        super(SkipAwareTestResult, self).addFailure(test, err)
        self._create_pdb(descr, "fail")

    def addSkip(self, test, reason):
        self.skipped.append((test, reason))
        if self.showAll:
            self.stream.writeln("SKIPPED")
        elif self.dots:
            self.stream.write("S")

    def printErrors(self) -> None:
        super(SkipAwareTestResult, self).printErrors()
        self.printSkippedList()

    def printSkippedList(self) -> None:
        # format (test, err) compatible with unittest2
        for test, err in self.skipped:
            descr = self.getDescription(test)
            self.stream.writeln(self.separator1)
            self.stream.writeln("%s: %s" % ("SKIPPED", descr))
            self.stream.writeln("\t%s" % err)

    def printErrorList(self, flavour, errors):
        for (_, descr), (test, err) in zip(self.descrs_for(flavour), errors):
            self.stream.writeln(self.separator1)
            self.stream.writeln("%s: %s" % (flavour, descr))
            self.stream.writeln(self.separator2)
            self.stream.writeln(err)
            self.stream.writeln("no stdout".center(len(self.separator2)))
            self.stream.writeln("no stderr".center(len(self.separator2)))


orig_call = testlib.TestCase.__call__


@monkeypatch(testlib.TestCase, "__call__")
def call(
    self: Any,
    result: SkipAwareTestResult = None,
    runcondition: Optional[Callable] = None,
    options: Optional[Any] = None,
) -> None:
    orig_call(self, result=result, runcondition=runcondition, options=options)
    # mypy: Item "None" of "Optional[Any]" has no attribute "exitfirst"
    # we check it first in the if
    if hasattr(options, "exitfirst") and options.exitfirst:  # type: ignore
        # add this test to restart file
        try:
            restartfile = open(FILE_RESTART, "a")
            try:
                descr = ".".join(
                    (self.__class__.__module__, self.__class__.__name__, self._testMethodName)
                )
                restartfile.write(descr + os.linesep)
            finally:
                restartfile.close()
        except Exception:
            print(
                "Error while saving succeeded test into",
                osp.join(os.getcwd(), FILE_RESTART),
                file=sys.__stderr__,
            )
            raise


@monkeypatch(testlib.TestCase)
def defaultTestResult(self):
    """return a new instance of the defaultTestResult"""
    return SkipAwareTestResult()


class NonStrictTestLoader(unittest.TestLoader):
    """
    Overrides default testloader to be able to omit classname when
    specifying tests to run on command line.

    For example, if the file test_foo.py contains ::

        class FooTC(TestCase):
            def test_foo1(self): # ...
            def test_foo2(self): # ...
            def test_bar1(self): # ...

        class BarTC(TestCase):
            def test_bar2(self): # ...

    'python test_foo.py' will run the 3 tests in FooTC
    'python test_foo.py FooTC' will run the 3 tests in FooTC
    'python test_foo.py test_foo' will run test_foo1 and test_foo2
    'python test_foo.py test_foo1' will run test_foo1
    'python test_foo.py test_bar' will run FooTC.test_bar1 and BarTC.test_bar2
    """

    def __init__(self) -> None:
        self.skipped_patterns = ()

    # some magic here to accept empty list by extending
    # and to provide callable capability
    def loadTestsFromNames(self, names: List[str], module: type = None) -> TestSuite:
        suites = []
        for name in names:
            suites.extend(self.loadTestsFromName(name, module))
        return self.suiteClass(suites)

    def _collect_tests(self, module: type) -> Dict[str, Tuple[type, List[str]]]:
        tests = {}
        for obj in vars(module).values():
            if isclass(obj) and issubclass(obj, unittest.TestCase):
                classname = obj.__name__
                if classname[0] == "_" or self._this_is_skipped(classname):
                    continue
                methodnames = []
                # obj is a TestCase class
                for attrname in dir(obj):
                    if attrname.startswith(self.testMethodPrefix):
                        attr = getattr(obj, attrname)
                        if callable(attr):
                            methodnames.append(attrname)
                # keep track of class (obj) for convenience
                tests[classname] = (obj, methodnames)
        return tests

    def loadTestsFromSuite(self, module, suitename):
        try:
            suite = getattr(module, suitename)()
        except AttributeError:
            return []
        assert hasattr(suite, "_tests"), "%s.%s is not a valid TestSuite" % (
            module.__name__,
            suitename,
        )
        # python2.3 does not implement __iter__ on suites, we need to return
        # _tests explicitly
        return suite._tests

    def loadTestsFromName(self, name, module=None):
        parts = name.split(".")
        if module is None or len(parts) > 2:
            # let the base class do its job here
            return [super(NonStrictTestLoader, self).loadTestsFromName(name)]
        tests = self._collect_tests(module)
        collected = []
        if len(parts) == 1:
            pattern = parts[0]
            if callable(getattr(module, pattern, None)) and pattern not in tests:
                # consider it as a suite
                return self.loadTestsFromSuite(module, pattern)
            if pattern in tests:
                # case python unittest_foo.py MyTestTC
                klass, methodnames = tests[pattern]
                for methodname in methodnames:
                    collected = [klass(methodname) for methodname in methodnames]
            else:
                # case python unittest_foo.py something
                for klass, methodnames in tests.values():
                    # skip methodname if matched by skipped_patterns
                    for skip_pattern in self.skipped_patterns:
                        methodnames = [
                            methodname
                            for methodname in methodnames
                            if skip_pattern not in methodname
                        ]
                    collected += [
                        klass(methodname) for methodname in methodnames if pattern in methodname
                    ]
        elif len(parts) == 2:
            # case "MyClass.test_1"
            classname, pattern = parts
            klass, methodnames = tests.get(classname, (None, []))
            for methodname in methodnames:
                collected = [
                    klass(methodname) for methodname in methodnames if pattern in methodname
                ]
        return collected

    def _this_is_skipped(self, testedname: str) -> bool:
        # mypy: Need type annotation for 'pat'
        # doc doesn't say how to that in list comprehension
        return any([(pat in testedname) for pat in self.skipped_patterns])  # type: ignore

    def getTestCaseNames(self, testCaseClass: type) -> List[str]:
        """Return a sorted sequence of method names found within testCaseClass"""
        is_skipped = self._this_is_skipped
        classname = testCaseClass.__name__
        if classname[0] == "_" or is_skipped(classname):
            return []
        testnames = super(NonStrictTestLoader, self).getTestCaseNames(testCaseClass)
        return [testname for testname in testnames if not is_skipped(testname)]


# The 2 functions below are modified versions of the TestSuite.run method
# that is provided with unittest2 for python 2.6, in unittest2/suite.py
# It is used to monkeypatch the original implementation to support
# extra runcondition and options arguments (see in testlib.py)


def _ts_run(
    self: Any,
    result: SkipAwareTestResult,
    debug: bool = False,
    runcondition: Callable = None,
    options: Optional[Any] = None,
) -> SkipAwareTestResult:
    self._wrapped_run(result, runcondition=runcondition, options=options)
    self._tearDownPreviousClass(None, result)
    self._handleModuleTearDown(result)
    return result


def _ts_wrapped_run(
    self: Any,
    result: SkipAwareTestResult,
    debug: bool = False,
    runcondition: Callable = None,
    options: Optional[Any] = None,
) -> SkipAwareTestResult:
    for test in self:
        if result.shouldStop:
            break
        if unittest_suite._isnotsuite(test):
            self._tearDownPreviousClass(test, result)
            self._handleModuleFixture(test, result)
            self._handleClassSetUp(test, result)
            result._previousTestClass = test.__class__
            if getattr(test.__class__, "_classSetupFailed", False) or getattr(
                result, "_moduleSetUpFailed", False
            ):
                continue

        # --- modifications to deal with _wrapped_run ---
        # original code is:
        #
        # if not debug:
        #     test(result)
        # else:
        #     test.debug()
        if hasattr(test, "_wrapped_run"):
            try:
                test._wrapped_run(result, debug, runcondition=runcondition, options=options)
            except TypeError:
                test._wrapped_run(result, debug)
        elif not debug:
            try:
                test(result, runcondition, options)
            except TypeError:
                test(result)
        else:
            test.debug()
        # --- end of modifications to deal with _wrapped_run ---
    return result


if sys.version_info >= (2, 7):
    # The function below implements a modified version of the
    # TestSuite.run method that is provided with python 2.7, in
    # unittest/suite.py
    def _ts_run(  # noqa
        self: Any,
        result: SkipAwareTestResult,
        debug: bool = False,
        runcondition: Callable = None,
        options: Optional[Any] = None,
    ) -> SkipAwareTestResult:
        topLevel = False
        if getattr(result, "_testRunEntered", False) is False:
            result._testRunEntered = topLevel = True

        self._wrapped_run(result, debug, runcondition, options)

        if topLevel:
            self._tearDownPreviousClass(None, result)
            self._handleModuleTearDown(result)
            result._testRunEntered = False
        return result


def enable_dbc(*args):
    """
    Without arguments, return True if contracts can be enabled and should be
    enabled (see option -d), return False otherwise.

    With arguments, return False if contracts can't or shouldn't be enabled,
    otherwise weave ContractAspect with items passed as arguments.
    """
    if not ENABLE_DBC:
        return False
    try:
        from logilab.aspects.weaver import weaver
        from logilab.aspects.lib.contracts import ContractAspect
    except ImportError:
        sys.stderr.write("Warning: logilab.aspects is not available. Contracts disabled.")
        return False
    for arg in args:
        weaver.weave_module(arg, ContractAspect)
    return True


# monkeypatch unittest and doctest (ouch !)
unittest._TextTestResult = SkipAwareTestResult
unittest.TextTestRunner = SkipAwareTextTestRunner
unittest.TestLoader = NonStrictTestLoader
unittest.TestProgram = SkipAwareTestProgram

if sys.version_info >= (2, 4):
    doctest.DocTestCase.__bases__ = (testlib.TestCase,)
    # XXX check python2.6 compatibility
    # doctest.DocTestCase._cleanups = []
    # doctest.DocTestCase._out = []
else:
    unittest.FunctionTestCase.__bases__ = (testlib.TestCase,)
unittest.TestSuite.run = _ts_run
unittest.TestSuite._wrapped_run = _ts_wrapped_run

if __name__ == "__main__":
    run()



# copyright 2003-2011 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of logilab-common.
#
# logilab-common is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option) any
# later version.
#
# logilab-common is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with logilab-common.  If not, see <http://www.gnu.org/licenses/>.
"""Cache module, with a least recently used algorithm for the management of the
deletion of entries.




"""
__docformat__ = "restructuredtext en"

from threading import Lock

from logilab.common.decorators import locked
from typing import TypeVar, List

_marker = object()


_KeyType = TypeVar("_KeyType")


class Cache(dict):
    """A dictionary like cache.

    inv:
        len(self._usage) <= self.size
        len(self.data) <= self.size
    """

    def __init__(self, size: int = 100) -> None:
        """Warning : Cache.__init__() != dict.__init__().
        Constructor does not take any arguments beside size.
        """
        assert size >= 0, "cache size must be >= 0 (0 meaning no caching)"
        self.size = size
        self._usage: List = []
        self._lock = Lock()
        super(Cache, self).__init__()

    def _acquire(self) -> None:
        self._lock.acquire()

    def _release(self) -> None:
        self._lock.release()

    def _update_usage(self, key: _KeyType) -> None:
        if not self._usage:
            self._usage.append(key)
        elif self._usage[-1] != key:
            try:
                self._usage.remove(key)
            except ValueError:
                # we are inserting a new key
                # check the size of the dictionary
                # and remove the oldest item in the cache
                if self.size and len(self._usage) >= self.size:
                    super(Cache, self).__delitem__(self._usage[0])
                    del self._usage[0]
            self._usage.append(key)
        else:
            pass  # key is already the most recently used key

    def __getitem__(self, key: _KeyType):
        value = super(Cache, self).__getitem__(key)
        self._update_usage(key)
        return value

    __getitem__ = locked(_acquire, _release)(__getitem__)

    def __setitem__(self, key: _KeyType, item):
        # Just make sure that size > 0 before inserting a new item in the cache
        if self.size > 0:
            super(Cache, self).__setitem__(key, item)
            self._update_usage(key)

    __setitem__ = locked(_acquire, _release)(__setitem__)

    def __delitem__(self, key: _KeyType):
        super(Cache, self).__delitem__(key)
        self._usage.remove(key)

    __delitem__ = locked(_acquire, _release)(__delitem__)

    def clear(self):
        super(Cache, self).clear()
        self._usage = []

    clear = locked(_acquire, _release)(clear)

    def pop(self, key: _KeyType, default=_marker):
        if key in self:
            self._usage.remove(key)
        # if default is _marker:
        #    return super(Cache, self).pop(key)
        return super(Cache, self).pop(key, default)

    pop = locked(_acquire, _release)(pop)

    def popitem(self):
        raise NotImplementedError()

    def setdefault(self, key, default=None):
        raise NotImplementedError()

    def update(self, other):
        raise NotImplementedError()



from __future__ import print_function

import logging
import urllib2

import kerberos as krb


import re

RGX = re.compile(r"(?:.*,)*\s*Negotiate\s*([^,]*),?", re.I)


class GssapiAuthError(Exception):
    """raised on error during authentication process"""


def get_negociate_value(headers):
    for authreq in headers.getheaders("www-authenticate"):
        match = RGX.search(authreq)
        if match:
            return match.group(1)


class HTTPGssapiAuthHandler(urllib2.BaseHandler):
    """Negotiate HTTP authentication using context from GSSAPI"""

    handler_order = 400  # before Digest Auth

    def __init__(self):
        self._reset()

    def _reset(self):
        self._retried = 0
        self._context = None

    def clean_context(self):
        if self._context is not None:
            krb.authGSSClientClean(self._context)

    def http_error_401(self, req, fp, code, msg, headers):
        try:
            if self._retried > 5:
                raise urllib2.HTTPError(
                    req.get_full_url(), 401, "negotiate auth failed", headers, None
                )
            self._retried += 1
            logging.debug("gssapi handler, try %s" % self._retried)
            negotiate = get_negociate_value(headers)
            if negotiate is None:
                logging.debug("no negociate found in a www-authenticate header")
                return None
            logging.debug("HTTPGssapiAuthHandler: negotiate 1 is %r" % negotiate)
            result, self._context = krb.authGSSClientInit("HTTP@%s" % req.get_host())
            if result < 1:
                raise GssapiAuthError("HTTPGssapiAuthHandler: init failed with %d" % result)
            result = krb.authGSSClientStep(self._context, negotiate)
            if result < 0:
                raise GssapiAuthError("HTTPGssapiAuthHandler: step 1 failed with %d" % result)
            client_response = krb.authGSSClientResponse(self._context)
            logging.debug("HTTPGssapiAuthHandler: client response is %s..." % client_response[:10])
            req.add_unredirected_header("Authorization", "Negotiate %s" % client_response)
            server_response = self.parent.open(req)
            negotiate = get_negociate_value(server_response.info())
            if negotiate is None:
                logging.warning("HTTPGssapiAuthHandler: failed to authenticate server")
            else:
                logging.debug("HTTPGssapiAuthHandler negotiate 2: %s" % negotiate)
                result = krb.authGSSClientStep(self._context, negotiate)
                if result < 1:
                    raise GssapiAuthError("HTTPGssapiAuthHandler: step 2 failed with %d" % result)
            return server_response
        except GssapiAuthError as exc:
            logging.error(repr(exc))
        finally:
            self.clean_context()
            self._reset()


if __name__ == "__main__":
    import sys

    # debug
    import httplib

    httplib.HTTPConnection.debuglevel = 1
    httplib.HTTPSConnection.debuglevel = 1

    logging.basicConfig(level=logging.DEBUG)
    # handle cookies
    import cookielib

    cj = cookielib.CookieJar()
    ch = urllib2.HTTPCookieProcessor(cj)
    # test with url sys.argv[1]
    h = HTTPGssapiAuthHandler()
    response = urllib2.build_opener(h, ch).open(sys.argv[1])
    print("\nresponse: %s\n--------------\n" % response.code, response.info())



# copyright 2003-2012 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of logilab-common.
#
# logilab-common is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option) any
# later version.
#
# logilab-common is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with logilab-common.  If not, see <http://www.gnu.org/licenses/>.
"""Date manipulation helper functions."""
from __future__ import division

__docformat__ = "restructuredtext en"

import math
import re
import sys
from locale import getlocale, LC_TIME
from datetime import date, time, datetime, timedelta
from time import strptime as time_strptime
from calendar import monthrange, timegm
from typing import Union, List, Any, Optional, Generator

from logilab.common.deprecation import callable_deprecated

try:
    from mx.DateTime import RelativeDateTime, Date, DateTimeType
except ImportError:
    endOfMonth = None
    DateTimeType = datetime
else:
    endOfMonth = RelativeDateTime(months=1, day=-1)

# NOTE: should we implement a compatibility layer between date representations
#       as we have in lgc.db ?

FRENCH_FIXED_HOLIDAYS = {
    "jour_an": "%s-01-01",
    "fete_travail": "%s-05-01",
    "armistice1945": "%s-05-08",
    "fete_nat": "%s-07-14",
    "assomption": "%s-08-15",
    "toussaint": "%s-11-01",
    "armistice1918": "%s-11-11",
    "noel": "%s-12-25",
}

FRENCH_MOBILE_HOLIDAYS = {
    "paques2004": "2004-04-12",
    "ascension2004": "2004-05-20",
    "pentecote2004": "2004-05-31",
    "paques2005": "2005-03-28",
    "ascension2005": "2005-05-05",
    "pentecote2005": "2005-05-16",
    "paques2006": "2006-04-17",
    "ascension2006": "2006-05-25",
    "pentecote2006": "2006-06-05",
    "paques2007": "2007-04-09",
    "ascension2007": "2007-05-17",
    "pentecote2007": "2007-05-28",
    "paques2008": "2008-03-24",
    "ascension2008": "2008-05-01",
    "pentecote2008": "2008-05-12",
    "paques2009": "2009-04-13",
    "ascension2009": "2009-05-21",
    "pentecote2009": "2009-06-01",
    "paques2010": "2010-04-05",
    "ascension2010": "2010-05-13",
    "pentecote2010": "2010-05-24",
    "paques2011": "2011-04-25",
    "ascension2011": "2011-06-02",
    "pentecote2011": "2011-06-13",
    "paques2012": "2012-04-09",
    "ascension2012": "2012-05-17",
    "pentecote2012": "2012-05-28",
}

# XXX this implementation cries for multimethod dispatching


def get_step(dateobj: Union[date, datetime], nbdays: int = 1) -> timedelta:
    # assume date is either a python datetime or a mx.DateTime object
    if isinstance(dateobj, date):
        return ONEDAY * nbdays
    return nbdays  # mx.DateTime is ok with integers


def datefactory(
    year: int, month: int, day: int, sampledate: Union[date, datetime]
) -> Union[date, datetime]:
    # assume date is either a python datetime or a mx.DateTime object
    if isinstance(sampledate, datetime):
        return datetime(year, month, day)
    if isinstance(sampledate, date):
        return date(year, month, day)
    return Date(year, month, day)


def weekday(dateobj: Union[date, datetime]) -> int:
    # assume date is either a python datetime or a mx.DateTime object
    if isinstance(dateobj, date):
        return dateobj.weekday()
    return dateobj.day_of_week


def str2date(datestr: str, sampledate: Union[date, datetime]) -> Union[date, datetime]:
    # NOTE: datetime.strptime is not an option until we drop py2.4 compat
    year, month, day = [int(chunk) for chunk in datestr.split("-")]
    return datefactory(year, month, day, sampledate)


def days_between(start: Union[date, datetime], end: Union[date, datetime]) -> int:
    if isinstance(start, date):
        # mypy: No overload variant of "__sub__" of "datetime" matches argument type "date"
        # we ensure that end is a date
        assert isinstance(end, date)
        delta = end - start  # type: ignore
        # datetime.timedelta.days is always an integer (floored)
        if delta.seconds:
            return delta.days + 1
        return delta.days
    else:
        return int(math.ceil((end - start).days))


def get_national_holidays(
    begin: Union[date, datetime], end: Union[date, datetime]
) -> Union[List[date], List[datetime]]:
    """return french national days off between begin and end"""
    begin = datefactory(begin.year, begin.month, begin.day, begin)
    end = datefactory(end.year, end.month, end.day, end)
    holidays = [str2date(datestr, begin) for datestr in FRENCH_MOBILE_HOLIDAYS.values()]
    for year in range(begin.year, end.year + 1):
        for datestr in FRENCH_FIXED_HOLIDAYS.values():
            date = str2date(datestr % year, begin)
            if date not in holidays:
                holidays.append(date)
    return [day for day in holidays if begin <= day < end]


def add_days_worked(start: date, days: int) -> date:
    """adds date but try to only take days worked into account"""
    step = get_step(start)
    weeks, plus = divmod(days, 5)
    end = start + ((weeks * 7) + plus) * step
    if weekday(end) >= 5:  # saturday or sunday
        end += 2 * step
    end += len([x for x in get_national_holidays(start, end + step) if weekday(x) < 5]) * step
    if weekday(end) >= 5:  # saturday or sunday
        end += 2 * step
    return end


def nb_open_days(start: Union[date, datetime], end: Union[date, datetime]) -> int:
    assert start <= end
    step = get_step(start)
    days = days_between(start, end)
    weeks, plus = divmod(days, 7)
    if weekday(start) > weekday(end):
        plus -= 2
    elif weekday(end) == 6:
        plus -= 1
    open_days = weeks * 5 + plus
    nb_week_holidays = len(
        [x for x in get_national_holidays(start, end + step) if weekday(x) < 5 and x < end]
    )
    open_days -= nb_week_holidays
    if open_days < 0:
        return 0
    return open_days


def date_range(
    begin: date, end: date, incday: Optional[Any] = None, incmonth: Optional[bool] = None
) -> Generator[date, Any, None]:
    """yields each date between begin and end

    :param begin: the start date
    :param end: the end date
    :param incr: the step to use to iterate over dates. Default is
                 one day.
    :param include: None (means no exclusion) or a function taking a
                    date as parameter, and returning True if the date
                    should be included.

    When using mx datetime, you should *NOT* use incmonth argument, use instead
    oneDay, oneHour, oneMinute, oneSecond, oneWeek or endOfMonth (to enumerate
    months) as `incday` argument
    """
    assert not (incday and incmonth)
    begin = todate(begin)
    end = todate(end)
    if incmonth:
        while begin < end:
            yield begin
            begin = next_month(begin, incmonth)
    else:
        incr = get_step(begin, incday or 1)
        while begin < end:
            yield begin
            begin += incr


# makes py datetime usable #####################################################

ONEDAY: timedelta = timedelta(days=1)
ONEWEEK: timedelta = timedelta(days=7)

strptime = callable_deprecated("Use strptime from datetime.datetime instead")(datetime.strptime)


def strptime_time(value, format="%H:%M"):
    return time(*time_strptime(value, format)[3:6])


def todate(somedate: date) -> date:
    """return a date from a date (leaving unchanged) or a datetime"""
    if isinstance(somedate, datetime):
        return date(somedate.year, somedate.month, somedate.day)
    assert isinstance(somedate, (date, DateTimeType)), repr(somedate)
    return somedate


def totime(somedate):
    """return a time from a time (leaving unchanged), date or datetime"""
    # XXX mx compat
    if not isinstance(somedate, time):
        return time(somedate.hour, somedate.minute, somedate.second)
    assert isinstance(somedate, (time)), repr(somedate)
    return somedate


def todatetime(somedate):
    """return a date from a date (leaving unchanged) or a datetime"""
    # take care, datetime is a subclass of date
    if isinstance(somedate, datetime):
        return somedate
    assert isinstance(somedate, (date, DateTimeType)), repr(somedate)
    return datetime(somedate.year, somedate.month, somedate.day)


def datetime2ticks(somedate: Union[date, datetime]) -> int:
    return timegm(somedate.timetuple()) * 1000 + int(getattr(somedate, "microsecond", 0) / 1000)


def ticks2datetime(ticks: int) -> datetime:
    miliseconds, microseconds = divmod(ticks, 1000)
    try:
        return datetime.fromtimestamp(miliseconds)
    except (ValueError, OverflowError):
        epoch = datetime.fromtimestamp(0)
        nb_days, seconds = divmod(int(miliseconds), 86400)
        delta = timedelta(nb_days, seconds=seconds, microseconds=microseconds)
        try:
            return epoch + delta
        except (ValueError, OverflowError):
            raise


def days_in_month(somedate: date) -> int:
    return monthrange(somedate.year, somedate.month)[1]


def days_in_year(somedate):
    feb = date(somedate.year, 2, 1)
    if days_in_month(feb) == 29:
        return 366
    else:
        return 365


def previous_month(somedate, nbmonth=1):
    while nbmonth:
        somedate = first_day(somedate) - ONEDAY
        nbmonth -= 1
    return somedate


def next_month(somedate: date, nbmonth: int = 1) -> date:
    while nbmonth:
        somedate = last_day(somedate) + ONEDAY
        nbmonth -= 1
    return somedate


def first_day(somedate):
    return date(somedate.year, somedate.month, 1)


def last_day(somedate: date) -> date:
    return date(somedate.year, somedate.month, days_in_month(somedate))


def ustrftime(somedate: datetime, fmt: str = "%Y-%m-%d") -> str:
    """like strftime, but returns a unicode string instead of an encoded
    string which may be problematic with localized date.
    """
    if sys.version_info >= (3, 3):
        # datetime.date.strftime() supports dates since year 1 in Python >=3.3.
        return somedate.strftime(fmt)
    else:
        try:
            if sys.version_info < (3, 0):
                encoding = getlocale(LC_TIME)[1] or "ascii"
                return str(somedate.strftime(str(fmt)), encoding)
            else:
                return somedate.strftime(fmt)
        except ValueError:
            if somedate.year >= 1900:
                raise
            # datetime is not happy with dates before 1900
            # we try to work around this, assuming a simple
            # format string
            fields = {
                "Y": somedate.year,
                "m": somedate.month,
                "d": somedate.day,
            }
            if isinstance(somedate, datetime):
                fields.update({"H": somedate.hour, "M": somedate.minute, "S": somedate.second})
            fmt = re.sub("%([YmdHMS])", r"%(\1)02d", fmt)
            return str(fmt) % fields


def utcdatetime(dt: datetime) -> datetime:
    if dt.tzinfo is None:
        return dt
    # mypy: No overload variant of "__sub__" of "datetime" matches argument type "None"
    return dt.replace(tzinfo=None) - dt.utcoffset()  # type: ignore


def utctime(dt):
    if dt.tzinfo is None:
        return dt
    return (dt + dt.utcoffset() + dt.dst()).replace(tzinfo=None)


def datetime_to_seconds(date):
    """return the number of seconds since the begining of the day for that date"""
    return date.second + 60 * date.minute + 3600 * date.hour


def timedelta_to_days(delta):
    """return the time delta as a number of seconds"""
    return delta.days + delta.seconds / (3600 * 24)


def timedelta_to_seconds(delta):
    """return the time delta as a fraction of days"""
    return delta.days * (3600 * 24) + delta.seconds
